//
//  AppDelegate.swift
//  FrequencyCounter
//
//  Created by Phidgets on 2018-02-22.
//  Copyright © 2018 Phidgets. All rights reserved.
//

import Cocoa
import Phidget22Swift

@NSApplicationMain
class AppDelegate: NSObject, NSApplicationDelegate {

    @IBOutlet weak var window: NSWindow!
    let ch = FrequencyCounter()
    @IBOutlet weak var phidgetInfoBox: PhidgetInfoBox!
    
    //GUI
    @IBOutlet weak var settingsBox: NSBox!
    @IBOutlet weak var enabledCheck: NSButton!
    @IBOutlet weak var filterTypePopup: NSPopUpButton!
    @IBOutlet weak var filterTypeBox: NSBox!
    @IBOutlet weak var frequencyCutoffText: NSTextField!
    @IBOutlet weak var dataIntervalSlider: NSSlider!
    @IBOutlet weak var minDataIntervalLabel: NSTextField!
    @IBOutlet weak var maxDataIntervalLabel: NSTextField!
    @IBOutlet weak var dataIntervalLabel: NSTextField!
    @IBOutlet weak var extraSettingsBox: NSBox!
    @IBOutlet weak var powerSupplyPopup: NSPopUpButton!
    @IBOutlet weak var inputModePopup: NSPopUpButton!
    @IBOutlet weak var dataBox: NSBox!
    @IBOutlet weak var countLabel: NSTextField!
    @IBOutlet weak var pulsesInLabel: NSTextField!
    @IBOutlet weak var frequencyLabel: NSTextField!
    @IBOutlet weak var totalCountLabel: NSTextField!

    
    //Error Window
    @IBOutlet weak var errorEventWindow: NSPanel!
    @IBOutlet var errorEventLog: NSTextView!
    @IBOutlet weak var errorEventCount: NSTextField!
    var errorCounter = 0

    //MARK: View events
    func applicationDidFinishLaunching(_ aNotification: Notification) {
        do {
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
            
            let _ = ch.attach.addHandler(attach_handler)
            let _ = ch.detach.addHandler(detach_handler)
            let _ = ch.error.addHandler(error_handler)
            let _ = ch.frequencyChange.addHandler(frequencychange_handler)
            let _ = ch.countChange.addHandler(countchange_handler)
            
            /*
             * Please review the Phidget22 channel matching documentation for details on the device
             * and class architecture of Phidget22, and how channels are matched to device features.
             */
            
            /*
             * Specifies the serial number of the device to attach to.
             * For VINT devices, this is the hub serial number.
             *
             * The default is any device.
             */
            // try ch.setDeviceSerialNumber(<YOUR DEVICE SERIAL NUMBER>)
            
            /*
             * For VINT devices, this specifies the port the VINT device must be plugged into.
             *
             * The default is any port.
             */
            // try ch.setHubPort(0)
            
            /*
             * Specifies which channel to attach to.  It is important that the channel of
             * the device is the same class as the channel that is being opened.
             *
             * The default is any channel.
             */
            // try ch.setChannel(0)
            
            /*
             * In order to attach to a network Phidget, the program must connect to a Phidget22 Network Server.
             * In a normal environment this can be done automatically by enabling server discovery, which
             * will cause the client to discovery and connect to available servers.
             *
             * To force the channel to only match a network Phidget, set remote to 1.
             */
            
            // try Net.enableServerDiscovery(serverType: .deviceRemote)
            // try ch.setIsRemote(true)
            
            /*
             openCmdLine allows command line arguments to be passed to this program in order to
             configure channel matching.
             If you have modified the code and added your own channel matching, comment out
             the openCmdLine call, and simply replace it with the following line:
             */
            //try ch.open()
            
            try phidgetInfoBox.openCmdLine(phid: ch as Phidget)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func applicationShouldTerminateAfterLastWindowClosed(_ sender: NSApplication) -> Bool {
        return true
    }
    
    func applicationWillTerminate(_ aNotification: Notification) {
        //Safely close example
        do{
            print("Closing")
            ch.attach.removeAllHandlers()
            ch.detach.removeAllHandlers()
            ch.error.removeAllHandlers()
            ch.frequencyChange.removeAllHandlers()
            ch.countChange.removeAllHandlers()
            try ch.close()
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Event handlers
    func attach_handler(sender: Phidget){
        do{
            let attachedDevice = sender as! FrequencyCounter
            
            let deviceID = try attachedDevice.getDeviceID()
            let minDataInterval = try attachedDevice.getMinDataInterval()
            let maxDataInterval = try attachedDevice.getMaxDataInterval()
            let dataInterval = try attachedDevice.getDataInterval()
            let enabled = try attachedDevice.getEnabled()
            let frequencyCutoff = try attachedDevice.getFrequencyCutoff()
            let filterType = try attachedDevice.getFilterType()
            var powerSupply:PowerSupply?
            var inputMode:InputMode?
            
            if(deviceID == DeviceID.PN_DAQ1400){
                powerSupply = try attachedDevice.getPowerSupply()
                inputMode = try attachedDevice.getInputMode()
            }
            
            DispatchQueue.main.async {
                self.phidgetInfoBox.fillPhidgetInfo(device: sender)
                self.dataBox.isHidden = false
                self.settingsBox.isHidden = false
                self.filterTypeBox.isHidden = false
                self.enabledCheck.isHidden = false

                
                self.minDataIntervalLabel.stringValue = String(format:"%d",minDataInterval)
                self.maxDataIntervalLabel.stringValue = String(format:"%d",maxDataInterval)
                self.dataIntervalLabel.stringValue = String(format:"%d ms",dataInterval)
                
                self.dataIntervalSlider.minValue = Double(minDataInterval)
                self.dataIntervalSlider.maxValue = Double(maxDataInterval)
                self.dataIntervalSlider.doubleValue = Double(dataInterval)
                
                self.filterTypePopup.addItem(withTitle: "Zero Crossing")
                self.filterTypePopup.addItem(withTitle: "Logic Level")
                self.filterTypePopup.selectItem(at: (filterType == FrequencyFilterType.zeroCrossing) ? 0 : 1)
                
                self.frequencyCutoffText.stringValue = String(format:"%.2f", frequencyCutoff)
                
                if(deviceID == DeviceID.PN_DAQ1400){
                    self.enabledCheck.isHidden = true
                    self.filterTypeBox.isHidden = true
                    self.extraSettingsBox.isHidden = false
                    self.powerSupplyPopup.addItem(withTitle: "Off")
                    self.powerSupplyPopup.addItem(withTitle: "12V")
                    self.powerSupplyPopup.addItem(withTitle: "24V")
                    self.inputModePopup.addItem(withTitle: "NPN")
                    self.inputModePopup.addItem(withTitle: "PNP")
                    self.powerSupplyPopup.selectItem(at: powerSupply == PowerSupply.off ? 0 : powerSupply == PowerSupply.volts_12 ? 1 : 2)
                    self.inputModePopup.selectItem(at: inputMode == InputMode.NPN ? 0 : 1)
                }
                
                //Adjusting window height
                var frame = self.window.frame
                frame.size.height = self.window.maxSize.height
                self.window.setFrame(frame, display: true, animate: false)
                self.window.center()
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func detach_handler(sender: Phidget){
        DispatchQueue.main.async {
            self.phidgetInfoBox.fillPhidgetInfo(device: nil)
            self.dataBox.isHidden = true
            self.settingsBox.isHidden = true
            //Adjusting window height
            var frame = self.window.frame
            frame.size.height = self.window.minSize.height
            self.window.setFrame(frame, display: true, animate: false)
            self.window.center()
        }
    }
    
    func error_handler(sender: Phidget, data: (code: ErrorEventCode, description: String)){
        outputError(errorDescription: data.description, errorCode: data.code.rawValue)
    }
    
    func frequencychange_handler(sender: FrequencyCounter, frequency: Double){
        do{
            let count = try ch.getCount()
            DispatchQueue.main.async {
                self.frequencyLabel.stringValue = String(format:"%.1f Hz", frequency)
                self.totalCountLabel.stringValue = String(format:"%lli",count)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    func countchange_handler(sender: FrequencyCounter, data: (counts: UInt64, timeChange: Double)){
        do{
            let count = try ch.getCount()
            DispatchQueue.main.async {
                self.countLabel.stringValue = String(format:"%llu",data.counts)
                self.pulsesInLabel.stringValue = String(format:"%.2f ms", data.timeChange)
                self.totalCountLabel.stringValue = String(format:"%lli",count)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: GUI Controls
    @IBAction func setEnabled(_ sender: Any) {
        do{
            let check = sender as! NSButton
            try ch.setEnabled((check.state == NSControl.StateValue.on) ? true : false)
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setFilterType(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setFilterType(FrequencyFilterType.zeroCrossing)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setFilterType(FrequencyFilterType.logicLevel)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    @IBAction func setPowerSupply(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setPowerSupply(PowerSupply.off)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setPowerSupply(PowerSupply.volts_12)
            }
            else{
                try ch.setPowerSupply(PowerSupply.volts_24)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setInputMode(_ sender: Any) {
        do{
            let popup = sender as! NSPopUpButton
            if(popup.indexOfSelectedItem == 0){
                try ch.setInputMode(InputMode.NPN)
            }
            else if(popup.indexOfSelectedItem == 1){
                try ch.setInputMode(InputMode.PNP)
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setDataInterval(_ sender: Any) {
        let slider = sender as! NSSlider
        do{
            try ch.setDataInterval(UInt32(slider.intValue))
            dataIntervalLabel.stringValue = String(format:"%d ms", UInt32(slider.intValue))
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func setFrequencyCutoff(_ sender: Any) {
        do{
            if let cutoff = Double(frequencyCutoffText.stringValue){
                try ch.setFrequencyCutoff(cutoff)
            }
            else{
                frequencyCutoffText.stringValue = "Invalid"
            }
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    @IBAction func reset(_ sender: Any) {
        do{
            try ch.reset()
            let count = try ch.getCount()
            totalCountLabel.stringValue = "\(count)"
        } catch let err as PhidgetError {
            outputError(errorDescription: err.description, errorCode: err.errorCode.rawValue)
        } catch {
            print (error)
        }
    }
    
    //MARK: Error
    func outputError(errorDescription:String, errorCode:UInt32){
        DispatchQueue.main.async {
            self.errorCounter += 1
            let outputString = NSAttributedString(string: String(format:"Error %d: %@\n", errorCode, errorDescription))
            self.errorEventLog.textStorage?.beginEditing()
            self.errorEventLog.textStorage?.append(outputString)
            self.errorEventLog.textStorage?.endEditing()
            self.errorEventCount.stringValue = "\(self.errorCounter)"
            if(!self.errorEventWindow.isVisible){
                self.errorEventWindow.setIsVisible(true)
                self.errorEventWindow.center()
            }
        }
        
    }
    
    @IBAction func clearError(_ sender: Any) {
        errorEventLog.textStorage?.setAttributedString(NSAttributedString(string: ""))
        errorCounter = 0
        errorEventCount.stringValue = "0"
    }

}

